/***************************************************************************
* File:          	ioctl.c
* Description:   	ioctl interface implementation
* Dependence:    	hptintf.h
***************************************************************************/
#include "gui_lib.c"

extern atomic_t DPC_Request_Nums;

/*--------------------------------------------------------------------------------------------------------- */
#pragma pack(1)

typedef struct _HPT_REBUILD_PARAM
{
	DEVICEID idMirror;
	DWORD Lba;
	UCHAR nSector;
} HPT_REBUILD_PARAM, *PHPT_REBUILD_PARAM;

#pragma pack()

#define MAX_EVENTS 10
static HPT_EVENT hpt_event_queue[MAX_EVENTS];
static int event_queue_head=0, event_queue_tail=0;

/*---------------------------------------------------------------------------------------------------------- */
static int event_queue_add(PHPT_EVENT pEvent)
{
	int p;

	p = (event_queue_tail + 1) % MAX_EVENTS;
	if (p==event_queue_head)
	{
		return -1;
	}
	hpt_event_queue[event_queue_tail] = *pEvent;
	event_queue_tail = p;
	return 0;
}

static int event_queue_remove(PHPT_EVENT pEvent)
{
	if (event_queue_head != event_queue_tail)
	{
		*pEvent = hpt_event_queue[event_queue_head];
		event_queue_head++;
		event_queue_head %= MAX_EVENTS;
		return 0;
	}
	return -1;
}

#ifdef BEEPER_ON_MB
static int use_beeper = 1;

#if defined(MODULE)
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,0)
module_param(use_beeper, uint, 0);
#else 
MODULE_PARM(use_beeper, "i");
#endif
MODULE_PARM_DESC(use_beeper, "enable on-card buzzer support");
#endif

struct timer_list beeper_timer;
int need_beep = 0;
static int beeping = 0;

static void beeper_timer_fn(unsigned long data)
{
	if (beeping) {
		beeping = 0;
		OutPort(0x61, InPort(0x61) & 0xFC);
	}
	else if (need_beep) {
		beeping = 1;
		OutPort(0x61, InPort(0x61) | 3);
		OutPort(0x43, 0xB6);
		OutPort(0x42, 500 & 0xFF); /* frequency 500 */
		OutPort(0x42, 500>>8);
	}
	if (need_beep) {
		beeper_timer.expires = jiffies + HZ*70/100;
		add_timer(&beeper_timer);
	}
}

void stop_beeper(void)
{
	need_beep = 0;
}
#endif

void HPTLIBAPI ioctl_ReportEvent(UCHAR event, PVOID param)
{
	HPT_EVENT e;
	ZeroMemory(&e, sizeof(e));
	e.EventType = event;
	switch(event)
	{
#ifdef SUPPORT_ARRAY
		case ET_INITIALIZE_ABORTED:
		case ET_INITIALIZE_FAILED:
			memcpy(e.Data, ((PVDevice)param)->u.array.ArrayName, MAX_ARRAY_NAME);
		case ET_INITIALIZE_STARTED:
		case ET_INITIALIZE_FINISHED:
		case ET_REBUILD_STARTED:
		case ET_REBUILD_ABORTED:
		case ET_REBUILD_FAILED:
		case ET_REBUILD_FINISHED:
		case ET_VERIFY_STARTED:
		case ET_VERIFY_ABORTED:
		case ET_VERIFY_FAILED:
		case ET_VERIFY_FINISHED:
		case ET_VERIFY_DATA_ERROR:
		case ET_SPARE_TOOK_OVER:
#endif
		case ET_DEVICE_REMOVED:
		case ET_DEVICE_PLUGGED:
			e.DeviceID = VDEV_TO_ID((PVDevice)param);
			break;

		default:
			break;
	}
	event_queue_add(&e);

#ifdef BEEPER_ON_MB
	if (event==ET_DEVICE_REMOVED) {
		if (use_beeper && need_beep==0) {
			need_beep = 1;
			beeper_timer.data = 0;
			beeper_timer.function = beeper_timer_fn;
			beeper_timer.expires = jiffies + HZ*70/100;
			add_timer(&beeper_timer);
		}
	}
#endif
}

static void HPTLIBAPI nothing(_VBUS_ARG void *notused){}

unsigned long lock_driver_idle(_VBUS_ARG0)
{
	unsigned long flags;

	spin_lock_irqsave(&driver_lock, flags);
	if (CURRCMDs) {
		hpt_printk(("CURRCMDs is %d, dpc %d, wait..\n", 
			CURRCMDs, atomic_read(&DPC_Request_Nums)));
		if (!mWaitingForIdle(_VBUS_P0)) CallWhenIdle(_VBUS_P nothing, 0);
		atomic_inc(&DPC_Request_Nums);
		do {
			spin_unlock_irqrestore(&driver_lock, flags);
			schedule();
			spin_lock_irqsave(&driver_lock, flags);
		}
		while (CURRCMDs);
	}
	else
		atomic_inc(&DPC_Request_Nums);
	CheckIdleCall(_VBUS_P0);
	return flags;
}

void unlock_driver_idle(_VBUS_ARG unsigned long flags)
{
	atomic_dec(&DPC_Request_Nums);
	Check_Idle_Call(_VBUS_P0);
	CheckPendingCall(_VBUS_P0);
	spin_unlock_irqrestore(&driver_lock, flags);
}


static int hpt_device_io(DEVICEID id, int cmd, ULONG lba, DWORD nSector, PVOID buffer);
static int hpt_get_event(PHPT_EVENT pEvent);
int hpt_set_array_state(DEVICEID idArray, DWORD state);
int hpt_rescan_all(void);

static void HPTLIBAPI pass_through_cmd_done(_VBUS_ARG PCommand pCmd)
{
	up(pCmd->pOrgCommand);
}

static void cmd_timeout(unsigned long data)
{
	PCommand pCmd = (PCommand)data;
	pCmd->Result = RETURN_SELECTION_TIMEOUT;
	up(pCmd->pOrgCommand);
}

static void timer_start_for_cmd(PCommand pCmd, int second)
{
	unsigned long timeout;
	struct timer_list timer;

	timeout = jiffies + second * HZ;
	init_timer(&timer);
	timer.expires = timeout;
	timer.data = (ULONG_PTR)pCmd;
	timer.function = cmd_timeout;
	add_timer(&timer);
	 
	if (down_interruptible(pCmd->pOrgCommand))
		down(pCmd->pOrgCommand);

	del_timer(&timer);
}


int Kernel_DeviceIoControl(
							DWORD dwIoControlCode,       	/* operation control code */
							PVOID lpInBuffer,            	/* input data buffer */
							DWORD nInBufferSize,         	/* size of input data buffer */
							PVOID lpOutBuffer,           	/* output data buffer */
							DWORD nOutBufferSize,        	/* size of output data buffer */
							PDWORD lpBytesReturned      	/* byte count */
						)
{
	unsigned long flags;
	
	switch(dwIoControlCode)
	{
#ifdef SUPPORT_ARRAY
        case HPT_IOCTL_DELETE_ARRAY:
		{
			DEVICEID idArray;
			int iSuccess;
	       	int i;
			PVDevice pArray;
			PVBus _vbus_p;

			if (nInBufferSize!=sizeof(DEVICEID)+sizeof(DWORD)) return -1;
			if (nOutBufferSize!=sizeof(int)) return -1;
			idArray = *(DEVICEID *)lpInBuffer;

			pArray = ID_TO_VDEV(idArray);

	        if((idArray==0) || check_VDevice_valid(pArray))	
		        return -1;
		
	        if(!mIsArray(pArray))
		        return -1;

		_vbus_p=pArray->pVBus;

		for(i = 0; i < MAX_VDEVICE_PER_VBUS; i++)
			if(pArray == _vbus_p->pVDevice[i]) {
				if (sd_inuse(hpt_vhost, i)) {
					hpt_printk(("Can not delete a mounted device.\n"));
					return -1;
				}
				break;
			} 

			flags = lock_driver_idle(_VBUS_P0);
			iSuccess = hpt_delete_array(_VBUS_P idArray, *(DWORD*)((DEVICEID *)lpInBuffer+1));
			unlock_driver_idle(_VBUS_P flags);

			*(int*)lpOutBuffer = iSuccess;

			if(iSuccess != 0)
				return -1;
			break;
		}
#endif
		case HPT_IOCTL_DEVICE_IO: /*reserve? */
		{
			PVOID buffer;
			DEVICEID id;
			int cmd;
			ULONG lba;
			DWORD nSector;
			int iSuccess;

			if (nInBufferSize < sizeof(HPT_DEVICE_IO)) return -1;

			id		=	((PHPT_DEVICE_IO)lpInBuffer)->id;
			cmd		=	((PHPT_DEVICE_IO)lpInBuffer)->cmd;
			lba		=	((PHPT_DEVICE_IO)lpInBuffer)->lba;
			nSector	=	((PHPT_DEVICE_IO)lpInBuffer)->nSector;
			buffer  =	((PHPT_DEVICE_IO)lpInBuffer)->buffer;

		    iSuccess = hpt_device_io(id, cmd, lba, nSector, buffer);

			if(cmd == IO_COMMAND_READ)
			{
				memcpy(lpOutBuffer, &iSuccess, sizeof(int));
				memcpy(lpOutBuffer+4, buffer, nSector*512);
			}
			else
			{
				*(int*)lpOutBuffer = iSuccess;
			}

			if(iSuccess != 0)
				return -1;
		}
		break;

		case HPT_IOCTL_IDE_PASS_THROUGH:
		{
			PVDevice pVDev;
			PCommand	pCmd;
			struct semaphore	sem;
			PIDE_PASS_THROUGH_HEADER in, out;

			if ( nInBufferSize<sizeof(IDE_PASS_THROUGH_HEADER)
			 || nOutBufferSize<sizeof(IDE_PASS_THROUGH_HEADER) )
				return -1;

			in = (PIDE_PASS_THROUGH_HEADER)lpInBuffer;
			out = (PIDE_PASS_THROUGH_HEADER)lpOutBuffer;

			pVDev = ID_TO_VDEV(in->idDisk);
			if (!pVDev || check_VDevice_valid(pVDev))	
		        	return -1;

			flags = lock_driver_idle(_VBUS_P0);

			pCmd = AllocateCommand(_VBUS_P0);
			if (!pCmd) {
				unlock_driver_idle(_VBUS_P flags);
				return -1;
			}

			pCmd->cf_ide_passthrough = 1;
			sema_init(&sem, 0);
			pCmd->pOrgCommand = &sem;
			pCmd->pVDevice = pVDev;
			pCmd->pfnCompletion = pass_through_cmd_done;

			if (in->protocol==IO_COMMAND_READ) {
				pCmd->cf_data_in = 1;
				pCmd->uCmd.Passthrough.pDataBuffer = IDE_PASS_THROUGH_buffer(out);
			}
			else if (in->protocol==IO_COMMAND_WRITE) {
				pCmd->cf_data_out = 1;
				pCmd->uCmd.Passthrough.pDataBuffer = IDE_PASS_THROUGH_buffer(in);
			}

			pCmd->uCmd.Passthrough.bCommandReg = in->bCommandReg;
			pCmd->uCmd.Passthrough.bLbaLowReg = in->bLbaLowReg;
			pCmd->uCmd.Passthrough.bLbaMidReg = in->bLbaMidReg;
			pCmd->uCmd.Passthrough.bLbaHighReg = in->bLbaHighReg;
			pCmd->uCmd.Passthrough.bDriveHeadReg = in->bDriveHeadReg;
			pCmd->uCmd.Passthrough.bFeaturesReg = in->bFeaturesReg;
			pCmd->uCmd.Passthrough.bSectorCountReg = in->bSectorCountReg;
			pCmd->uCmd.Passthrough.nSectors = in->nSectors;

			pVDev->pfnSendCommand(_VBUS_P pCmd);
			CheckPendingCall(_VBUS_P0);

			unlock_driver_idle(_VBUS_P flags);

			timer_start_for_cmd(pCmd, 10);

			spin_lock_irqsave(&driver_lock, flags);

			if (pCmd->Result != RETURN_SUCCESS) {
				fResetVBus(_VBUS_P0);
				FreeCommand(_VBUS_P pCmd);
				spin_unlock_irqrestore(&driver_lock, flags);
				return -1;
			}

			out->bCommandReg = pCmd->uCmd.Passthrough.bCommandReg;
			out->bLbaLowReg = pCmd->uCmd.Passthrough.bLbaLowReg;
			out->bLbaMidReg = pCmd->uCmd.Passthrough.bLbaMidReg;
			out->bLbaHighReg = pCmd->uCmd.Passthrough.bLbaHighReg;
			out->bDriveHeadReg = pCmd->uCmd.Passthrough.bDriveHeadReg;
			out->bFeaturesReg = pCmd->uCmd.Passthrough.bFeaturesReg;
			out->bSectorCountReg = pCmd->uCmd.Passthrough.bSectorCountReg;
			out->nSectors = pCmd->uCmd.Passthrough.nSectors;

			FreeCommand(_VBUS_P pCmd);
			spin_unlock_irqrestore(&driver_lock, flags);
		}
		break;
		
		case HPT_IOCTL_GET_EVENT:
		{
			PHPT_EVENT pInfo;

			if (nInBufferSize!=0) return -1;
			if (nOutBufferSize!=sizeof(HPT_EVENT)) return -1;

			/* check auto rebuild */
			if (autorebuild) {
				spin_lock_irqsave(&driver_lock, flags);
				fRescanAllDevice(1);
				if ((CURRCMDs == 0) && (atomic_read(&DPC_Request_Nums) == 0))
					Check_Idle_Call();
				spin_unlock_irqrestore(&driver_lock, flags);
			}

			pInfo = (PHPT_EVENT)lpOutBuffer;

			if (hpt_get_event(pInfo)!=0)
				return -1;
		}
		break;
#ifdef SUPPORT_ARRAY
        case HPT_IOCTL_REBUILD_DATA_BLOCK:
			return -1;

		case HPT_IOCTL_SET_ARRAY_STATE:
		{
			DEVICEID idArray;
			DWORD state;

			if (nInBufferSize!=sizeof(HPT_SET_STATE_PARAM)) return -1;
			if (nOutBufferSize!=0) return -1;

			idArray = ((PHPT_SET_STATE_PARAM)lpInBuffer)->idArray;
			state = ((PHPT_SET_STATE_PARAM)lpInBuffer)->state;

			if(hpt_set_array_state(idArray, state)!=0)
				return -1;
		}
		break;
#endif
		case HPT_IOCTL_RESCAN_DEVICES:
		{
			if (nInBufferSize!=0) return -1;
			if (nOutBufferSize!=0) return -1;
			hpt_rescan_all();
		}
		break;

		case HPT_IOCTL_LOCK_DEVICE:
			return -1;
	
		case HPT_IOCTL_UNLOCK_DEVICE:
			return -1;

		case HPT_IOCTL_SET_DEVICE_INFO:
			if (((PHPT_SET_DEVICE_INFO)lpInBuffer)->idDisk==0) {
				return 0;
			}
			return hpt_default_ioctl(_VBUS_P dwIoControlCode, lpInBuffer, nInBufferSize, 
				lpOutBuffer, nOutBufferSize, lpBytesReturned);

		default:
		{
			if(hpt_default_ioctl(dwIoControlCode, lpInBuffer, nInBufferSize, 
				lpOutBuffer, nOutBufferSize, lpBytesReturned) == -1) return -1;
#ifdef SUPPORT_ARRAY
			switch(dwIoControlCode)
			{
				case HPT_IOCTL_CREATE_ARRAY:
					spin_lock_irqsave(&driver_lock, flags);
                    if(((PCREATE_ARRAY_PARAMS)lpInBuffer)->CreateFlags & CAF_CREATE_AND_DUPLICATE)
				    {
						  (ID_TO_VDEV(*(DEVICEID *)lpOutBuffer))->u.array.rf_auto_rebuild = 0;
                          hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, ID_TO_VDEV(*(DEVICEID *)lpOutBuffer), DUPLICATE);
					}
					else if(((PCREATE_ARRAY_PARAMS)lpInBuffer)->CreateFlags & CAF_CREATE_R5_ZERO_INIT)
				    {
                          hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, ID_TO_VDEV(*(DEVICEID *)lpOutBuffer), INITIALIZE);
					}
					else if(((PCREATE_ARRAY_PARAMS)lpInBuffer)->CreateFlags & CAF_CREATE_R5_BUILD_PARITY)
				    {
                          hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, ID_TO_VDEV(*(DEVICEID *)lpOutBuffer), REBUILD_PARITY);
					}
					spin_unlock_irqrestore(&driver_lock, flags);
                    break;

	
				case HPT_IOCTL_ADD_DISK_TO_ARRAY:
				{
					PVDevice pArray = ID_TO_VDEV(((PHPT_ADD_DISK_TO_ARRAY)lpInBuffer)->idArray);
					if(pArray->u.array.rf_rebuilding == 0)
					{		
						unsigned long timeout;
						timeout = jiffies + 3 * HZ;	
						spin_lock_irqsave(&driver_lock, flags);
						pArray->u.array.rf_auto_rebuild = 0;
						pArray->u.array.rf_abort_rebuild = 0;
						hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, pArray, DUPLICATE);
						spin_unlock_irqrestore(&driver_lock, flags);
						while (!pArray->u.array.rf_rebuilding && 
							time_before(jiffies, timeout)) schedule();
					}
					break;
				}
			}
#endif
            return 0;
		}
	}

	if (lpBytesReturned)
		*lpBytesReturned = nOutBufferSize;
	return 0;
}

/*---------------------------------------------------------------------------------------------------------- */
int hpt_device_io(DEVICEID id, int cmd, ULONG lba, DWORD nSector, PVOID buffer)
{
	return -1;
}

int hpt_get_event(PHPT_EVENT pEvent)
{
	return event_queue_remove(pEvent);
}

#ifdef SUPPORT_ARRAY
int hpt_set_array_state(DEVICEID idArray, DWORD state)
{
	PVDevice pVDevice = ID_TO_VDEV(idArray);
	int	i;
	unsigned long timeout;
	unsigned long flags;

	if (idArray==0 || check_VDevice_valid(pVDevice))	return -1;
	if(!mIsArray(pVDevice)) return -1;
	if(!pVDevice->vf_online || pVDevice->u.array.rf_broken) return -1;

    timeout = jiffies + 80 * HZ; /* should be long than hard-reset time */
	switch(state)
	{
		case MIRROR_REBUILD_START:
		{
			if (pVDevice->u.array.rf_rebuilding ||
				pVDevice->u.array.rf_verifying ||
				pVDevice->u.array.rf_initializing)
				return -1;
			
			spin_lock_irqsave(&driver_lock, flags);

			pVDevice->u.array.rf_auto_rebuild = 0;
			pVDevice->u.array.rf_abort_rebuild = 0;
            hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, pVDevice, 
				(UCHAR)((pVDevice->u.array.CriticalMembers || pVDevice->VDeviceType == VD_RAID_1)? DUPLICATE : REBUILD_PARITY));

			spin_unlock_irqrestore(&driver_lock, flags);

			while (!pVDevice->u.array.rf_rebuilding && 
				time_before(jiffies, timeout)) schedule();
		}

		break;

		case AS_VERIFY_START:
		{
			if (pVDevice->u.array.rf_rebuilding ||
				pVDevice->u.array.rf_verifying ||
				pVDevice->u.array.rf_initializing)
				return -1;
	
			spin_lock_irqsave(&driver_lock, flags);
            
            pVDevice->u.array.RebuildSectors = 0;
			hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, pVDevice, VERIFY_R);

			spin_unlock_irqrestore(&driver_lock, flags);
			
			while (!(pVDevice->u.array.rf_verifying || pVDevice->u.array.rf_rebuilding)
				    && time_before(jiffies, timeout)) schedule();
		}
		break;

		case AS_INITIALIZE_START:
		{
			if (pVDevice->u.array.rf_rebuilding ||
				pVDevice->u.array.rf_verifying ||
				pVDevice->u.array.rf_initializing)
				return -1;

			spin_lock_irqsave(&driver_lock, flags);
			hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, pVDevice, VERIFY_R);
			spin_unlock_irqrestore(&driver_lock, flags);
			
			while (!pVDevice->u.array.rf_initializing && 
				time_before(jiffies, timeout)) schedule();
		}
		break;

		case MIRROR_REBUILD_ABORT:
		{
			for(i = 0; i < pVDevice->u.array.bArnMember; i++) {
				if (pVDevice->u.array.pMember[i] && 
					mIsArray(pVDevice->u.array.pMember[i]) &&
					pVDevice->u.array.pMember[i]->u.array.rf_rebuilding)
				{
					/* take care if you use spinlock in recursive call */
					hpt_set_array_state(VDEV_TO_ID(pVDevice->u.array.pMember[i]), state);
				}
			}
			
			if(pVDevice->u.array.rf_rebuilding != 1)
				return -1;
			
			spin_lock_irqsave(&driver_lock, flags);
			pVDevice->u.array.rf_abort_rebuild = 1;
			spin_unlock_irqrestore(&driver_lock, flags);
			
			while (pVDevice->u.array.rf_rebuilding && 
				time_before(jiffies, timeout)) schedule();
		}
		break;

		case AS_VERIFY_ABORT:
		{
			if(pVDevice->u.array.rf_verifying != 1)
				return -1;

			spin_lock_irqsave(&driver_lock, flags);
			pVDevice->u.array.rf_abort_rebuild = 1;
			spin_unlock_irqrestore(&driver_lock, flags);
			
			while (pVDevice->u.array.rf_verifying && 
				time_before(jiffies, timeout)) schedule();
		}
		break;

		case AS_INITIALIZE_ABORT:
		{
			if(pVDevice->u.array.rf_initializing != 1)
				return -1;

			spin_lock_irqsave(&driver_lock, flags);
			pVDevice->u.array.rf_abort_rebuild = 1;
			spin_unlock_irqrestore(&driver_lock, flags);
			
			while (pVDevice->u.array.rf_initializing && 
				time_before(jiffies, timeout)) schedule();
		}
		break;

		default:
			return -1;
	}

	return 0;
}
#endif

void Check_Idle_Call(void);

void HPTLIBAPI do_rescan(_VBUS_ARG void *unused)
{
	fRescanAllDevice(0);
}

int hpt_rescan_all()
{
    unsigned long flags;
	unsigned long timeout;
    
    spin_lock_irqsave(&driver_lock, flags);
    CallWhenIdle(_VBUS_P do_rescan, 0);
    spin_unlock_irqrestore(&driver_lock, flags);

	/* wait the rescan to complete? */
	timeout = jiffies + 60 * HZ;
    while((CURRCMDs || atomic_read(&DPC_Request_Nums)) && time_before(jiffies, timeout)) {
    	schedule();
    	if (!mWaitingForIdle(gVBus)) return 0;
    }

	spin_lock_irqsave(&driver_lock, flags);
	if((CURRCMDs == 0) && (atomic_read(&DPC_Request_Nums) == 0))
	    Check_Idle_Call();
	spin_unlock_irqrestore(&driver_lock, flags);

    return 0;
}

/***************************************************************************
 * Function:     hpt_rebuild_data_block
 * Description:  rebuild a mirror block
 * Returns:      None
 ***************************************************************************/
#ifdef SUPPORT_ARRAY
int HPTLIBAPI R1ControlSgl(_VBUS_ARG PCommand pCmd, FPSCAT_GATH pSgTable, int logical)
{
	ULONG bufferSize = SECTOR_TO_BYTE(pCmd->uCmd.R1Control.nSectors);
	if (pCmd->uCmd.R1Control.Command==CTRL_CMD_VERIFY)
		bufferSize<<=1;
	if (logical) {
		pSgTable->dSgAddress = (ULONG_PTR)pCmd->uCmd.R1Control.Buffer;
		pSgTable->wSgSize = (USHORT)bufferSize;
		pSgTable->wSgFlag = SG_FLAG_EOT;
	}
	else {
		/* build physical SG table for pCmd->uCmd.R1Control.Buffer */
		pSgTable->dSgAddress = fOsPhysicalAddress(pCmd->uCmd.R1Control.Buffer);
		if(pSgTable->dSgAddress & 1) return 0;
        pSgTable->wSgSize = (USHORT)bufferSize;
		pSgTable->wSgFlag = SG_FLAG_EOT;
	}
	return 1;
}

static int End_Job = 0;

static void HPTLIBAPI thread_io_done(_VBUS_ARG PCommand pCmd)
{
	End_Job = 1;
	up((struct semaphore *)pCmd->pOrgCommand);
}

void hpt_rebuild_data_block(PVDevice pArray, UCHAR flags)
{
	ULONG capacity = pArray->VDeviceCapacity / (pArray->u.array.bArnMember-1);
   	PCommand pCmd;
	UINT result;
	int needsync=0, retry=0, needdelete=0;
	unsigned long Lock_Flag;
	void *buffer = 0;
	struct semaphore sem;

	_VBUS_INST(pArray->pVBus)

	if (pArray->u.array.RebuildSectors>=capacity)
		return;

	spin_lock_irqsave(&driver_lock, Lock_Flag);

	if (pArray->u.array.rf_broken) goto cmd_fail;
	
	switch(flags) {
	case DUPLICATE:
	case REBUILD_PARITY:
		if(pArray->u.array.rf_rebuilding == 0){
			pArray->u.array.rf_rebuilding = 1;
			hpt_printk(("Rebuilding started.\n"));
			ioctl_ReportEvent(ET_REBUILD_STARTED, pArray);
		}
		break;

	case INITIALIZE:
		if(pArray->u.array.rf_initializing == 0){
			pArray->u.array.rf_initializing = 1;
			hpt_printk(("Initializing started.\n"));
			ioctl_ReportEvent(ET_INITIALIZE_STARTED, pArray);
		}
		break;
	case VERIFY_R:
		if(pArray->u.array.rf_verifying == 0){
			pArray->u.array.rf_verifying = 1;
			hpt_printk(("Verifying started.\n"));
			ioctl_ReportEvent(ET_VERIFY_STARTED, pArray);
		}
		break;
	}

	KdPrintI(("rebuild_begin\n"));

retry_cmd:
	pCmd = AllocateCommand(_VBUS_P0);
	pCmd->cf_control = 1;
	End_Job = 0;
	sema_init(&sem, 0);
	pCmd->pOrgCommand = &sem;

	if (pArray->VDeviceType==VD_RAID_1) {
		/* 32K rebuild. */

		#define MAX_REBUILD_SECTORS 0x40
		if(!(buffer = kmalloc(SECTOR_TO_BYTE(MAX_REBUILD_SECTORS), GFP_ATOMIC))) {
			FreeCommand(pCmd);
			spin_unlock_irqrestore(&driver_lock, Lock_Flag);
			return;
		}
#ifdef SUPPORT_OLD_ARRAY
		if(pArray->vf_format_v2 == 1 && pArray->u.array.RebuildSectors < OLD_RAID_INFO_LBA) {
			switch(flags) {
			case INITIALIZE:
			case DUPLICATE:
				pCmd->uCmd.R1Control.Command = CTRL_CMD_REBUILD;
				pCmd->uCmd.R1Control.nSectors = (OLD_RAID_INFO_LBA-pArray->u.array.RebuildSectors);
				break;
			case VERIFY_R:
				pCmd->uCmd.R1Control.Command = CTRL_CMD_VERIFY;
				pCmd->uCmd.R1Control.nSectors = (OLD_RAID_INFO_LBA-pArray->u.array.RebuildSectors);
				break;

			}
		}	
		else 
#endif
		{
			switch(flags) {
			case INITIALIZE:				
			case DUPLICATE:
				pCmd->uCmd.R1Control.Command = CTRL_CMD_REBUILD;
				pCmd->uCmd.R1Control.nSectors = MAX_REBUILD_SECTORS;
				break;
			case VERIFY_R:
				pCmd->uCmd.R1Control.Command = CTRL_CMD_VERIFY;
				pCmd->uCmd.R1Control.nSectors = MAX_REBUILD_SECTORS/2;
				break;
			}
		}
#ifdef SUPPORT_OLD_ARRAY
		if (pArray->vf_format_v2 == 1 && pArray->u.array.RebuildSectors == OLD_RAID_INFO_LBA) pArray->u.array.RebuildSectors++;
#endif
		pCmd->uCmd.R1Control.Lba = pArray->u.array.RebuildSectors;
		if (capacity - pArray->u.array.RebuildSectors < pCmd->uCmd.R1Control.nSectors)
			pCmd->uCmd.R1Control.nSectors = capacity - pArray->u.array.RebuildSectors;

		pCmd->uCmd.R1Control.Buffer = buffer;
		pCmd->pfnBuildSgl = R1ControlSgl;
	}
	
	pCmd->pVDevice = pArray;
	pCmd->pfnCompletion = thread_io_done;
	pArray->pfnSendCommand(_VBUS_P pCmd);
	CheckPendingCall();

	spin_unlock_irqrestore(&driver_lock, Lock_Flag);
	
	if (!End_Job) {
		timer_start_for_cmd(pCmd, 60);
	}
	
	spin_lock_irqsave(&driver_lock, Lock_Flag);
	if (!End_Job) {
		hpt_printk(("timeout, reset!(%d)\n",pCmd->Result));
		fResetVBus(_VBUS_P0);
	}
	result = pCmd->Result;
	FreeCommand(_VBUS_P pCmd);
	if (buffer) kfree(buffer);
	KdPrintI(("cmd finished %d", result));

	switch(result) {
	case RETURN_SUCCESS:
		if (!pArray->u.array.rf_abort_rebuild) {
			if(pArray->u.array.RebuildSectors < capacity) {
				hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, pArray, flags);
			}
			else {
				switch (flags) {
				case DUPLICATE:
				case REBUILD_PARITY:
					needsync = 1;
					pArray->u.array.rf_rebuilding = 0;
					pArray->u.array.rf_need_rebuild = 0;
					pArray->u.array.CriticalMembers = 0;
					pArray->u.array.RebuildSectors = 0x7FFFFFFF;
					pArray->u.array.rf_duplicate_and_create = 0;
					hpt_printk(("Rebuilding finished.\n"));
					ioctl_ReportEvent(ET_REBUILD_FINISHED, pArray);
					break;
				case INITIALIZE:
					needsync = 1;
					pArray->u.array.rf_initializing = 0;
					pArray->u.array.rf_need_rebuild = 0;
					pArray->u.array.RebuildSectors = 0x7FFFFFFF;
					hpt_printk(("Initializing finished.\n"));
					ioctl_ReportEvent(ET_INITIALIZE_FINISHED, pArray);
					break;
				case VERIFY_R:
					pArray->u.array.rf_verifying = 0;
					hpt_printk(("Verifying finished.\n"));
					ioctl_ReportEvent(ET_VERIFY_FINISHED, pArray);
					break;
				}
			}
		}
		else {
			pArray->u.array.rf_abort_rebuild = 0;
			if (pArray->u.array.rf_rebuilding) {
				hpt_printk(("Abort rebuilding.\n"));
				pArray->u.array.rf_rebuilding = 0;
				pArray->u.array.rf_duplicate_and_create = 0;
				ioctl_ReportEvent(ET_REBUILD_ABORTED, pArray);
			}
			else if (pArray->u.array.rf_verifying) {
				hpt_printk(("Abort verifying.\n"));
				pArray->u.array.rf_verifying = 0;
				ioctl_ReportEvent(ET_VERIFY_ABORTED, pArray);
			}
			else if (pArray->u.array.rf_initializing) {
				hpt_printk(("Abort initializing.\n"));
				pArray->u.array.rf_initializing = 0;
				ioctl_ReportEvent(ET_INITIALIZE_ABORTED, pArray);
			}
			needdelete=1;
		}
		break;
	
	case RETURN_DATA_ERROR:
		if (flags==VERIFY_R) {
			/*
			 * verify may fail due to OS update I/O buffer asynchronously. 
			 * try to recover this case.
			 */
			if (pArray->VDeviceType==VD_RAID_1) {
				if (++retry<3) {
					spin_unlock_irqrestore(&driver_lock, Lock_Flag);
					set_current_state(TASK_INTERRUPTIBLE);
					schedule_timeout(10*HZ);
					spin_lock_irqsave(&driver_lock, Lock_Flag);
					goto retry_cmd;
				}
			}
			needsync = 1;
			pArray->u.array.rf_verifying = 0;
			pArray->u.array.rf_need_rebuild = 1;
			hpt_printk(("Verifying failed: found inconsistency\n"));
			ioctl_ReportEvent(ET_VERIFY_DATA_ERROR, pArray);
			ioctl_ReportEvent(ET_VERIFY_FAILED, pArray);

			if (!pArray->vf_online || pArray->u.array.rf_broken) break;

			pArray->u.array.rf_auto_rebuild = 0;
			pArray->u.array.rf_abort_rebuild = 0;
			hpt_queue_dpc((HPT_DPC)hpt_rebuild_data_block, pArray,
				(pArray->VDeviceType == VD_RAID_1) ? DUPLICATE : REBUILD_PARITY);
		}
		break;
		
	default:
		hpt_printk(("command failed \n"));
		if (++retry<3 && result != RETURN_BAD_DEVICE) {
			hpt_printk(("retry (%d)\n", retry));
			goto retry_cmd;
		}
cmd_fail:
		pArray->u.array.rf_abort_rebuild = 0;
		switch (flags) {
		case DUPLICATE:
		case REBUILD_PARITY:
			needsync = 1;
			pArray->u.array.rf_rebuilding = 0;
			pArray->u.array.rf_duplicate_and_create = 0;
			hpt_printk(((flags==DUPLICATE)? "Duplicating failed.\n":"Rebuilding failed.\n"));
			ioctl_ReportEvent(ET_REBUILD_FAILED, pArray);
			break;
		case INITIALIZE:
			needsync = 1;
			pArray->u.array.rf_initializing = 0;
			hpt_printk(("Initializing failed.\n"));
			ioctl_ReportEvent(ET_INITIALIZE_FAILED, pArray);
			break;
		case VERIFY_R:
			needsync = 1;
			pArray->u.array.rf_verifying = 0;
#ifdef SUPPORT_OLD_ARRAY
			if(pArray->vf_format_v2 == 1) pArray->u.array.RebuildSectors = MAX_LBA_T;
#endif
			hpt_printk(("Verifying failed.\n"));
			ioctl_ReportEvent(ET_VERIFY_FAILED, pArray);
			break;
		}
		needdelete=1;
	}			

	/* wait until no other active cmds. */
	while (CURRCMDs) {
		KdPrintI(("currcmds is %d, wait...\n", CURRCMDs));
		if (!mWaitingForIdle(_VBUS_P0)) CallWhenIdle(_VBUS_P nothing, 0);
		spin_unlock_irqrestore(&driver_lock, Lock_Flag);
		schedule();
		spin_lock_irqsave(&driver_lock, Lock_Flag);
	}

	if (needsync) SyncArrayInfo(pArray);
	if(needdelete && (pArray->u.array.rf_duplicate_must_done || (flags == INITIALIZE))) {
		CheckIdleCall(_VBUS_P0);
		fDeleteArray(pArray, TRUE);
	}

	if (atomic_read(&DPC_Request_Nums) < 1) /* see also lock_driver_idle */
	{
		Check_Idle_Call(_VBUS_P0);
		CheckPendingCall(_VBUS_P0);
	}

    spin_unlock_irqrestore(&driver_lock, Lock_Flag);
}
#endif


/****************************************************************
 *	Name:   hpt3xx_scsi_ioctl
 *	Description:	Process a scsi ioctl from the application, just like the S.M.A.R.T app.
 *	Parameters:		dev - Pointer to SCSI dev, which will be accessed, such sda...
 *					cmd - ioctl cmd type.
 					arg - buffer from app.
 *	Returns:		Status code.
 ****************************************************************/
int hpt3xx_scsi_ioctl(Scsi_Device *dev, int cmd, void *arg)
{
	unsigned char *buff=NULL;
	unsigned short copydata=0, offset=0;
	int err;
	DWORD dwRet;
	IDE_PASS_THROUGH_HEADER ide_passthrough_header;
	HPT_IOCTL_PARAM	ioctl_param;
	PVDevice pVDev = gVBus[dev->channel].pVDevice[dev->id];

	_VBUS_INST(pVBus);

	if (!pVDev || !pVDev->vf_online) {
		KdPrint(("ERROR: Device %d is offline. \n", dev->id));
		return -EINVAL;
	}

	/*We only support S.M.A.R.T fucntion on single disk, not array*/
	if (pVDev->VDeviceType != VD_SINGLE_DISK)  {
		KdPrint(("ERROR: Device %d is a RAID Array, Not supported. \n", dev->id));
		return -EINVAL;
	}

	/* Send the ioctl of the ide_pass_through_command */
	memset(&ioctl_param, 0, sizeof(ioctl_param));

	ide_passthrough_header.idDisk=VDEV_TO_ID(pVDev);
	
	if ( copy_from_user(&ide_passthrough_header.bFeaturesReg, arg+2, 1)
		||copy_from_user(&ide_passthrough_header.bSectorCountReg, arg+3, 1)
		||copy_from_user(&ide_passthrough_header.bLbaLowReg, arg+1, 1)
		||copy_from_user(&ide_passthrough_header.bCommandReg, arg, 1) )
		goto invalid;
	ide_passthrough_header.bLbaMidReg=0x4f;
	ide_passthrough_header.bLbaHighReg=0xc2;
	ide_passthrough_header.nSectors=1;
	ide_passthrough_header.protocol=0;
	ide_passthrough_header.bDriveHeadReg=pVDev->u.disk.bDeUnitId;

	switch ( cmd ) {
	/* CMD: HDIO_GET_IDENTITY */
	case 0x030d:
		offset=0;
		ide_passthrough_header.bCommandReg = 0xec;
		ide_passthrough_header.protocol=IO_COMMAND_READ;
		ide_passthrough_header.bSectorCountReg=1;
		copydata=512;
		ioctl_param.nInBufferSize=sizeof(IDE_PASS_THROUGH_HEADER);
		ioctl_param.nOutBufferSize=copydata+sizeof(IDE_PASS_THROUGH_HEADER);
	break;
	/* CMD: HDIO_DRIVER_CMD */
	case 0x031f:
		offset=4;
		switch ( ide_passthrough_header.bCommandReg ) {
		/* CHECK_POWER_MODE */
		case 0xe5 :
			ide_passthrough_header.protocol=IO_COMMAND_READ;
			/* Though smart monitor only copies one byte, error appears here if set to 1 */
			copydata=512;
		break;
		/* IDENTITY */
		case 0xec :
			ide_passthrough_header.protocol=IO_COMMAND_READ;
			copydata=512;
		break;
		/* PIDENTITY */
		case 0xa1 :
			ide_passthrough_header.protocol=IO_COMMAND_READ;
			copydata=512;
		break;
		/* ATA_SMART_COMMAND */
		case 0xb0:
			switch ( ide_passthrough_header.bFeaturesReg ) {
			/* READ_VALUES */
			case 0xd0 :
				ide_passthrough_header.protocol=IO_COMMAND_READ;
				copydata=512;
			break;
			/* READ_THRESHOLDS */
			case 0xd1 :
				ide_passthrough_header.protocol=IO_COMMAND_READ;
				copydata=512;
			break;
			/* READ_LOG */
			case 0xd5 :
				ide_passthrough_header.protocol=IO_COMMAND_READ;
				copydata=512;
			break;
			/* STATUS, ENABLE, DISABLE, AUTO_OFFLINE, AUTO_SAVE,
			IMMEDIATE_OFFLINE */
			}
		break;
		}
		ioctl_param.nInBufferSize=sizeof(IDE_PASS_THROUGH_HEADER);
		ioctl_param.nOutBufferSize=copydata+sizeof(IDE_PASS_THROUGH_HEADER);
	break;
	/* HDIO_DRIVER_TASK */
	case 0x031e:
		offset=7;
		if ( copy_from_user(&ide_passthrough_header.bFeaturesReg, arg+1, 1) )
			goto invalid;
		switch ( ide_passthrough_header.bFeaturesReg ){
		/* STATUS_CHECK */
		case 0xda:
			ioctl_param.nInBufferSize=sizeof(IDE_PASS_THROUGH_HEADER);
			ioctl_param.nOutBufferSize=sizeof(IDE_PASS_THROUGH_HEADER);
		break;
		}
	break;
	/* CMD: HDIO_DRIVER_TASKFILE */
	case 0x031d:
		if ( copy_from_user(&ide_passthrough_header.bCommandReg, arg+7, 1)
			||copy_from_user(&ide_passthrough_header.bSectorCountReg, arg+2, 1)
			||copy_from_user(&ide_passthrough_header.bLbaLowReg, arg+3, 1)
			||copy_from_user(&ide_passthrough_header.bDriveHeadReg, arg+6, 1)
			||copy_from_user(&ide_passthrough_header.bFeaturesReg, arg+1, 1) )
			goto invalid;
		ide_passthrough_header.protocol=IO_COMMAND_WRITE;
		switch ( ide_passthrough_header.bCommandReg ){
			/* SMART_ATA_COMMAND */
			case 0xb0 :
				switch ( ide_passthrough_header.bFeaturesReg ){
				/* WRITELOG */
				case 0xd6:
					copydata=512;
					ioctl_param.nInBufferSize=copydata+sizeof(IDE_PASS_THROUGH_HEADER);
					ioctl_param.nOutBufferSize=sizeof(IDE_PASS_THROUGH_HEADER);
				break;
				}
			break;
		}
	break;
	default:
		KdPrint(("ERROR: Unknown command %x!\n",cmd));
		return -EINVAL;
	}

	buff = kmalloc(copydata+2*sizeof(IDE_PASS_THROUGH_HEADER),GFP_ATOMIC);
	if ( !buff ) {
		hpt_printk(("ERROR: No sufficient memory! \n"));
		return -EINVAL;
	}
	memset(buff, 0, copydata+2*sizeof(IDE_PASS_THROUGH_HEADER));

	ioctl_param.dwIoControlCode=HPT_IOCTL_IDE_PASS_THROUGH;
	ioctl_param.lpInBuffer= buff;
	ioctl_param.lpOutBuffer=buff+sizeof(IDE_PASS_THROUGH_HEADER);

	if ( ide_passthrough_header.protocol==IO_COMMAND_WRITE && copydata )
		if ((hpt_verify_area(VERIFY_READ, (void *)((ULONG_PTR)arg+40), copydata)) ||
				(copy_from_user(buff+sizeof(IDE_PASS_THROUGH_HEADER), arg+40, copydata))) {
			KdPrint(("Got bad user address, read user address fail. \n"));
			goto invalid;
		}
	memcpy(buff, &ide_passthrough_header, sizeof(IDE_PASS_THROUGH_HEADER));

	err = Kernel_DeviceIoControl(_VBUS_P ioctl_param.dwIoControlCode, ioctl_param.lpInBuffer,
		ioctl_param.nInBufferSize, ioctl_param.lpOutBuffer, ioctl_param.nOutBufferSize, &dwRet);

	if (err) {
		KdPrint(("ERROR: hpt_scsi_ioctl return failed. \n"));
		goto invalid;
	}

	switch (cmd){
	case 0x031d:
		if( copy_to_user(arg, buff+sizeof(IDE_PASS_THROUGH_HEADER)+7, 1)
			||copy_to_user(arg+1, buff+sizeof(IDE_PASS_THROUGH_HEADER)+1, 6) )
			goto invalid;
	break;
	case 0x031f:
		if( copy_to_user(arg+1, buff+sizeof(IDE_PASS_THROUGH_HEADER)+1, 7) )
			goto invalid;
	break;
	default:
		if( copy_to_user(arg, buff+sizeof(IDE_PASS_THROUGH_HEADER)+7, 1)
			||copy_to_user(arg+1, buff+sizeof(IDE_PASS_THROUGH_HEADER)+3, 1)
			||copy_to_user(arg+2, buff+sizeof(IDE_PASS_THROUGH_HEADER)+2, 1) )
			goto invalid;
		/* according to Documentation/ioctl/hdio.txt (doc/hdio.txt) arg[3] undefined while ouput*/
	break;
	}
	if( ide_passthrough_header.protocol==IO_COMMAND_READ && copydata ){
		if((hpt_verify_area(VERIFY_WRITE, (void * )(ULONG_PTR)(arg+offset), copydata)) ||
			(copy_to_user(arg+offset,buff+2*sizeof(IDE_PASS_THROUGH_HEADER), copydata))) {
			KdPrint(("Got bad user address, write user address  fail. \n"));
			goto invalid;
		}
	}
	kfree(buff);
	return err;
invalid:
	kfree(buff);
	return -EINVAL;
}
